/* PrismJS 1.30.0 theme based on prism-solarized-light */
code[class*="language-"],
pre[class*="language-"] {
    color: var(--surfaceTxtHintColor);
    font-family: var(--monospaceFontFamily);
    font-size: 1em;
    text-align: left;
    white-space: pre;
    word-spacing: normal;
    word-break: normal;
    word-wrap: normal;
    line-height: 1.5;
    tab-size: 4;
    hyphens: none;
}
pre[class*="language-"] {
    padding: 1em;
    margin: 0.5em 0;
    overflow: auto;
}
.token.cdata,
.token.comment,
.token.doctype,
.token.prolog {
    color: var(--prismCommentColor);
}
.token.punctuation {
    color: var(--prismPunctuationColor);
}
.token.namespace {
    opacity: 0.7;
}
.token.constant,
.token.boolean,
.token.number {
    color: var(--prismNumberColor);
}
.token.deleted,
.token.property,
.token.symbol,
.token.tag,
.token.tag .token.punctuation {
    color: var(--prismPropertyColor);
}
.token.attr-name,
.token.builtin,
.token.char,
.token.inserted,
.token.selector,
.token.string,
.token.url {
    color: var(--prismStringColor);
}
.token.atrule,
.token.attr-value,
.token.keyword {
    color: var(--prismKeywordColor);
}
.token.class-name,
.token.function,
.token.important,
.token.regex,
.token.variable {
    color: var(--prismFunctionColor);
}
.token.bold {
    font-weight: bold;
}
.token.italic {
    font-style: italic;
}
.token.entity {
    cursor: help;
}
