.rate-limit-table {
    background: none;
    tr,
    td,
    th {
        background: none;
        padding: 0;
        border: 0;
        min-height: 0;
        height: auto;
    }
    th {
        padding: 0 0 10px;
        font-size: var(--smFontSize);
        line-height: var(--smLineHeight);
    }
    .rate-limit-row {
        .fields,
        .field,
        input,
        select,
        .input,
        .select {
            background: none;
            height: 100%;
            border-radius: 0;
        }
        td {
            background: var(--inputColor);
            border-left: 1px solid var(--inputBorderColor);
            border-top: 1px solid var(--inputBorderColor);
            transition: background var(--animationSpeed);
            &:first-child {
                border-left: 0px;
            }
            &:focus-within {
                background: var(--inputFocusColor);
            }
            &:has(.error:not(.hidden)) {
                background: var(--surfaceDangerColor);
            }
        }
        &:first-child {
            td {
                border-top: 0px;
            }
            td:first-child {
                border-top-left-radius: var(--borderRadius);
            }
            td:last-child {
                border-top-right-radius: var(--borderRadius);
            }
        }
        &:last-child {
            td:first-child {
                border-bottom-left-radius: var(--borderRadius);
            }
            td:last-child {
                border-bottom-right-radius: var(--borderRadius);
            }
        }
    }

    td.col-new-btn {
        padding: 3px;
        &:focus-within {
            background: var(--inputColor);
        }
    }
    .col-label {
        width: 60%;
    }
    .col-requests,
    .col-duration {
        width: 15%;
    }
    .col-audience {
        width: 1px;
        min-width: 130px;
        white-space: nowrap;
        .selected-container {
            flex-wrap: nowrap;
        }
    }
    .col-action {
        width: 1px;
        min-width: 0;
        padding: 0 3px;
    }
}

.rate-limit-table-wrapper {
    overflow: auto;
    .rate-limit-table {
        min-width: 550px;
    }
}
