.pb-tooltip {
    position: fixed;
    z-index: 99999;
    max-width: 300px;
    min-width: 30px;
    max-height: 90%;
    text-align: center;
    white-space: pre-line;
    word-break: break-word;
    background: var(--tooltipSurfaceColor);
    border-radius: var(--borderRadius);
    color: var(--tooltipTxtColor);
    font-weight: normal;
    font-size: var(--smFontSize);
    line-height: var(--smLineHeight);
    font-family: var(--baseFontFamily);
    scrollbar-width: thin;
    padding: 3px 5px;
    pointer-events: none;
    opacity: 0;
    transition:
        opacity var(--animationSpeed),
        display var(--animationSpeed) allow-discrete;
    &:popover-open {
        display: block;
        opacity: 1;
        @starting-style {
            opacity: 0;
        }
    }
    &[data-style="code"] {
        text-align: left;
        white-space: pre-wrap;
        max-width: 350px;
        font-family: var(--monospaceFontFamily);
    }
}
